/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Matthew Wilson <matthew@mjwilson.demon.co.uk>.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Doug Daniels <rainking@alumni.rice.edu> -- added more features for use in
 *                                               AnnotationsEngine
 *                                            -- XPCOM-ized
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/******************************
 * XPCOM Component Constants
 ******************************/
const NS_TRACKER_SERVICE_CONTRACTID =
    "@mozilla.org/tracker;1";
const NS_TRACKER_SERVICE_CID =
    Components.ID("{0ab8428e-90ac-444e-ae53-8dc1ec9ac19e}");
const NS_TRACKER_SERVICE_IID = Components.interfaces.nsITrackerService;
const nsISupports        = Components.interfaces.nsISupports;

/**
 * Tracker module for AnnotationsEngine.  Observes the locations
 * to which the user drives the browser and reports them to registered
 * observers.  Reports when a new page begins loading and when the
 * page finishes loading.
 */
function Tracker() {
    this.observers = [];
    this.state = null;
}

/**************
 * FIELDS
 **************/
Tracker.UNKNOWN_REASON_URL = 'about:blank';  

// internal states
Tracker.PAGE_URL_BLANK = 0;
Tracker.PAGE_LOADING = 1;
Tracker.PAGE_FINISHED = 2;

Tracker.NS_OBSERVER_SERVICE = "@mozilla.org/observer-service;1";

Tracker.prototype.observers = undefined;  // array of observers
Tracker.prototype.state = undefined;

Tracker.prototype.QueryInterface = function (iid) {
	
    if (!iid.equals(nsISupports) &&
	!iid.equals(NS_TRACKER_SERVICE_IID))
    throw Components.results.NS_ERROR_NO_INTERFACE;

    return this;
}


/**
 * Adds an observer to the list of
 * observers.  These observers receive updates
 * whenever a page changes.  Observers must implement
 * the interface nsITrackerObserver.
 * If adding our first observer, this function
 * turns on the tracking service.
 *
 * @param observer new observer to add
 */
Tracker.prototype.addObserver = function(observer) {
    
    this.observers.push(observer);
    
    if (this.observers.length == 1) {
	this.setEnabled(true);
    }
}

/**
 * Removes an observer from the list of observers.
 * If removing the last observer, we turn off the
 * tracking service.
 *
 * @param observer existing observer to remove
 */
Tracker.prototype.removeObserver = function(observer) {
    for (var i=0; i < this.observers.length; i++) {
	if (this.observers[i] == observer) {
	    // splice out this observer
	    this.observers.splice(i, 1);

	    // if we have no remaining observers, turn
	    // off the tracking service
	    if (this.observers.length == 0) {
		this.setEnabled(false);
	    }
	    return;
	}
    }

    // couldn't find the observer -- throw an error
    throw new Error("Tracker.RemoveObserver:  the observer " +
		    "requested for removal is not " +
		    "registered as an Observer.");
}



/**
 * observe will be called when there is a notification for the
 * topic |aTopic|.  This is an implementation of a method from
 * nsIObserver.idl.
 *
 * @param oSubject : Notification specific interface pointer -- the window this applies to
 * @param sMessage   : The notification topic or subject -- which event happened
 * @param sContextUrl : the url of the subject event.
 */
Tracker.prototype.observe = function (oSubject, sMessage, sContextUrl) {

    // get the window this applies to
    var win = oSubject.QueryInterface(Components.interfaces.nsIDOMWindow);

    // starting a new document load -- can load the annotations from 
    // the server
    if (sMessage == 'StartDocumentLoad') {
	this.state = Tracker.PAGE_LOADING;
	this.sendBeginLoad(Tracker.strip_URL(sContextUrl), win);
    }
    // document finalized, ok to insert icons
    else if (sMessage == 'EndDocumentLoad') {

	/* if we started the tracker in the middle of a page
	   loading, and didn't send the begin notice */
	if (this.state != Tracker.PAGE_LOADING) {
	    this.sendBeginLoad( Tracker.strip_URL(sContextUrl), win);
	}

	this.state = Tracker.PAGE_FINISHED;
	this.sendEndLoad(Tracker.strip_URL(sContextUrl), win);
    }
    else if (sMessage == 'FailDocumentLoad') {
	
	/* if we started the tracker in the middle of a page
	   loading, and didn't send the begin notice */
	if (this.state != Tracker.PAGE_LOADING) {
	    this.sendBeginLoad( Tracker.strip_URL(sContextUrl), win);
	}

	this.state = Tracker.PAGE_FINISHED;
	this.sendFailLoad( Tracker.strip_URL(sContextUrl), win);
    }
}

/**
 * Set the tracker enabled or disabled.
 *
 * @param enable whether to enable the tracker
 * @return whether the operation worked
 */
Tracker.prototype.setEnabled = function (enable) {
    if ( enable ) {
	// Install our navigation observer so we can track the main client window.
	var observerService = Components.classes[Tracker.NS_OBSERVER_SERVICE].getService();
	observerService = observerService.QueryInterface(Components.interfaces.nsIObserverService);
	if (observerService) {
	    observerService.addObserver(this, "StartDocumentLoad", false);
	    observerService.addObserver(this, "EndDocumentLoad", false);			
	    observerService.addObserver(this, "FailDocumentLoad", false);
	}
	else {
	    dump("nsITracker: FAILURE to get observer service\n");
	}
	
    }
    // unloading
    else {
	this.state = null;

	// remove our navigation observer.	
	var observerService = Components.classes[Tracker.NS_OBSERVER_SERVICE].getService();
	observerService = observerService.QueryInterface(Components.interfaces.nsIObserverService);
	if (observerService) {
	    observerService.removeObserver(this, "StartDocumentLoad");
	    observerService.removeObserver(this, "EndDocumentLoad");
	    observerService.removeObserver(this, "FailDocumentLoad");

	} else {
	    dump("nsITracker: FAILURE to get observer service--unable to unload observer\n");
	}
    }
}

/**
 * Sends notification to all registered
 * observers that a new page has begun
 * loading.
 *
 * @param url String location of new page
 */
Tracker.prototype.sendBeginLoad = function(url, win) {
    for (var i = 0; i < this.observers.length; i++) {
	this.observers[i].onBeginLoad(url, win);
    }
}

Tracker.prototype.sendEndLoad = function(url, win) {
    for (var i = 0; i < this.observers.length; i++) {
	this.observers[i].onEndLoad(url, win);
    }
}

Tracker.prototype.sendFailLoad = function(url, win) {
    for (var i = 0; i < this.observers.length; i++) {
	this.observers[i].onFailLoad(url, win);
    }
}
    

/**
 * Private function to strip query 
 * strings off of URLs. (we don't want
 * to be nosy!)
 *
 * @param url String URL to strip query strings from
 * @return URL string stripped of queries
 */
Tracker.strip_URL = function (url) {
    if (!url || ('string' != typeof(url)) ||  ('' == url) || url == 'about:blank') {
	return Tracker.UNKNOWN_REASON_URL;
    }
	
    // Strip off any query strings (Don't want to be too nosy).
    var queryPart = url.indexOf('?');
    if (queryPart != -1) {
	url = url.slice(0, queryPart);
    }
	
    return url;    
}

Tracker.prototype.toString = function () {
    return "[Tracker observers==" + this.observers.length + "]";
}


/**************************
 * XPCOM stuff
 **************************/

/**
 * Factory for nsTrackerService.
 */
var nsTrackerServiceFactory = new Object();

nsTrackerServiceFactory.createInstance = function (outer, iid) {
    if (outer != null) {
	throw Components.results.NS_ERROR_NO_AGGREGATION;
    }
    
    if (!iid.equals(nsISupports)) {
	throw Components.results.NS_ERROR_INVALID_ARG;
    }

    return new Tracker();
}


/**
 * Module for nsTrackerService.
 */
var nsTrackerServiceModule = new Object();

nsTrackerServiceModule.registerSelf =
    function (compMgr, fileSpec, location, type) {
	compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	dump("***** Tracker Service -- registering self ******\n");
	debug("***** Tracker Service -- registering self ******\n");
	compMgr.registerFactoryLocation(NS_TRACKER_SERVICE_CID,
					"Tracker Service",
					NS_TRACKER_SERVICE_CONTRACTID, 
					fileSpec,
					location,
					type);
    }

nsTrackerServiceModule.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.unregisterFactoryLocation(NS_TRACKER_SERVICE_CID, fileSpec);
}

nsTrackerServiceModule.getClassObject = function (compMgr, cid, iid) {
    if (cid.equals(NS_TRACKER_SERVICE_CID)) {
	return nsTrackerServiceFactory;
    }
    else {
	throw Components.results.NS_ERROR_NO_INTERFACE;  
    }
}

nsTrackerServiceModule.canUnload = function(compMgr) {
    return true;
}

/* entrypoint */
function NSGetModule(compMgr, fileSpec) {
    return nsTrackerServiceModule;
}

